MODULE crs
   !!===================================================================
   !!                  ***  crs.F90 ***
   !!  Purpose: Interface for calculating quantities from a  
   !!           higher-resolution grid for the coarse grid.
   !!
   !!  Method:  Given the user-defined reduction factor, 
   !!           the averaging bins are set: 
   !!           - cn_binref = NORTH, starting from the north 
   !!           to the south in the model interior domain, 
   !!           in this way the north fold and redundant halo cells  
   !!           could be handled in a consistent manner and 
   !!           the irregularities of bin size can be handled
   !!           more naturally by the presence of land
   !!           in the southern boundary.  Thus the southernmost bin 
   !!           could be of an irregular bin size.
   !!           Information on the parent grid is retained, specifically,
   !!           each coarse grid cell's volume and ocean surface 
   !!           at the faces, relative to the parent grid.
   !!           - cn_binref = EQUAT (not yet available), starting
   !!           at a centralized bin at the equator, being only
   !!           truly centered for odd-numbered j-direction reduction 
   !!           factors.
   !!  References:  Aumont, O., J.C. Orr, D. Jamous, P. Monfray
   !!               O. Marti and G. Madec, 1998. A degradation 
   !!               approach to accelerate simulations to steady-state 
   !!               in a 3-D tracer transport model of the global ocean.
   !!               Climate Dynamics, 14:101-116.
   !!  History: 
   !!       Original.   May 2012.  (J. Simeon, G. Madec, C. Ethe)
   !!===================================================================

   USE dom_oce        ! ocean space and time domain and to get jperio
!   USE wrk_nemo       ! work arrays
   USE crs_dom        ! domain for coarse grid
   USE in_out_manager 
   USE par_kind, ONLY: wp
   USE crslbclnk
   

   IMPLICIT NONE

   PRIVATE 

   PUBLIC crsfun, crs_e3_max, crs_surf

   INTERFACE crsfun
      MODULE PROCEDURE crsfun_mask, crsfun_coordinates, crsfun_wgt, crsfun_UV, crsfun_TW
   END INTERFACE

   !! Substitutions
#  include "domzgr_substitute.h90"
   
CONTAINS

   SUBROUTINE crsfun_mask( p_ptmask, cd_type, p_pmask2d, p_cmask, p_cmask2d )
      !!----------------------------------------------------------------
      !!               *** SUBROUTINE crsfun_mask ***
      !! ** Purpose :  Computes the 3D tmask of the coarse grid 
      !!
      !! ** Method  :  The subset of the parent grid tmask comprising
      !!               the new, coarsened grid box is examined for
      !!               the presence of an ocean mask point.  If, at
      !!               minimum, one ocean mask point is found, the 
      !!               new coarsened grid box becomes set to ocean (tmask=1)
      !!               The tmask is first determined, then from 
      !!               the coarse-grid tmask, the umask, vmask and fmask are 
      !!               derived.
      !! ** Input   : p_ptmask = 3D parent grid tmask
      !!              cd_type = grid type ('T', 'U', 'V', 'F')
      !!              p_pmask2d = (Optional) parent 2D mask (i.e. runoff mask)
      !! ** Output  : p_cmask   = (Optional) 3D coarsened grid [t|u|v|f]mask
      !!              p_cmask2d = (Optional) 2D coarsened grid [t|u|v|f]mask
      !! History.   29 May 2012.  2nd draft done.  Do u,v,f masks here or in crsini?
      !!            31 May 2012.  Redid umask,vmask, fmask creation to take into account blocked
      !!                          faces.               
      !!----------------------------------------------------------------
      ! Arguments
      REAL, DIMENSION(jpi,jpj,jpk)                 , INTENT(in)  :: p_ptmask  ! Parent grid tmask
      CHARACTER(len=1)                             , INTENT(in)  :: cd_type   ! grid type (T,U,V,F)
      REAL, DIMENSION(:,:),   OPTIONAL, ALLOCATABLE, INTENT(in)  :: p_pmask2d ! 2D mask, such as rnfmsk       
      REAL, DIMENSION(:,:,:), OPTIONAL, ALLOCATABLE, INTENT(out) :: p_cmask   ! Coarse grid [t,u,v,f]mask
      REAL, DIMENSION(:,:),   OPTIONAL, ALLOCATABLE, INTENT(out) :: p_cmask2d ! Coarse grid [t,u,v,f]mask


      ! Local variables
      INTEGER  ::  ji, jj, jk, ijpk                   ! dummy loop indices
      INTEGER  :: ijie,ijis,ijje,ijjs

      ! Initialize
      IF ( PRESENT( p_cmask ) ) THEN
         ALLOCATE( p_cmask(jpi_crs,jpj_crs,jpk) )
         p_cmask(:,:,:) = 0.0
         ijpk = jpk
      ENDIF

      IF ( PRESENT( p_cmask2d ) ) THEN
         ALLOCATE( p_cmask2d(jpi_crs,jpj_crs) )
         p_cmask2d(:,:) = 0.0
         ijpk = 1
      ENDIF
 

      DO jk = 1, ijpk
         DO ji = 2, jpi_crsm1
            ijie = mie_crs(ji)
            ijis = mis_crs(ji)

            DO jj = 2, jpj_crsm1
               ijje = mje_crs(jj) 
               ijjs = mjs_crs(jj)                   

               IF ( PRESENT( p_cmask ) ) THEN
               
                  IF ( cd_type == 'T' ) THEN  
                     p_cmask(ji,jj,jk) = SUM( p_ptmask(ijis:ijie,ijjs:ijje,jk) ) 
                     IF ( p_cmask(ji,jj,jk) > 0 )  p_cmask(ji,jj,jk) = 1
                  ELSEIF ( cd_type == 'V' ) THEN               
                     p_cmask(ji,jj,jk) = SUM( p_ptmask(ijis:ijie,ijje,jk) * p_ptmask(ijis:ijie,ijje+1,jk) )
                     IF ( p_cmask(ji,jj,jk) > 0 ) p_cmask(ji,jj,jk) = 1
                  ELSEIF ( cd_type == 'U' ) THEN
                     p_cmask(ji,jj,jk) = SUM( p_ptmask(ijie,ijjs:ijje,jk) * p_ptmask(ijie+1,ijjs:ijje,jk) ) 
                     IF ( p_cmask(ji,jj,jk) > 0 ) p_cmask(ji,jj,jk) = 1
                  ELSE   ! fmask
                     p_cmask(ji,jj,jk) =  p_ptmask(ijie,ijje,jk) + p_ptmask(ijie+1,ijje,jk)        &
                         &            + p_ptmask(ijie,ijje+1,jk) + p_ptmask(ijie+1,ijje+1,jk) 
                     IF ( p_cmask(ji,jj,jk) > 0 ) p_cmask(ji,jj,jk) = 1
                  ENDIF

               ENDIF

               IF ( PRESENT( p_cmask2d ) ) THEN
                  IF ( cd_type == 'T' ) THEN  
                     p_cmask2d(ji,jj) = SUM( p_pmask2d(ijis:ijie,ijjs:ijje) ) 
                     IF ( p_cmask2d(ji,jj) > 0 )  p_cmask2d(ji,jj) = 1
                  ELSEIF ( cd_type == 'V' ) THEN               
                     p_cmask2d(ji,jj) = SUM( p_pmask2d(ijis:ijie,ijje) * p_pmask2d(ijis:ijie,ijje+1) )
                     IF ( p_cmask2d(ji,jj) > 0 ) p_cmask2d(ji,jj) = 1
                  ELSEIF ( cd_type == 'U' ) THEN
                     p_cmask2d(ji,jj) = SUM( p_pmask2d(ijie,ijjs:ijje) * p_pmask2d(ijie+1,ijjs:ijje) ) 
                     IF ( p_cmask2d(ji,jj) > 0 ) p_cmask2d(ji,jj) = 1
                  ELSE   ! fmask
                     p_cmask2d(ji,jj) =  p_pmask2d(ijie,ijje) + p_pmask2d(ijie+1,ijje)        &
                         &            + p_pmask2d(ijie,ijje+1) + p_pmask2d(ijie+1,ijje+1) 
                     IF ( p_cmask2d(ji,jj) > 0 ) p_cmask2d(ji,jj) = 1
                  ENDIF

               ENDIF

            ENDDO
         ENDDO
      ENDDO


! Retroactively add back the halo cells j=1, j=jpj_crs and i=1, i=jpi_crs.

      IF ( PRESENT( p_cmask ) ) THEN
         IF( nperio /= 0 ) THEN
            CALL crs_lbc_lnk( cd_type,1.0,pt3d1=p_cmask )
         ELSE
            p_cmask( 1  ,:   ,:) = p_cmask(     2,:   ,:)          ! all points
            p_cmask(jpi_crs,:,:) = p_cmask(jpi_crsm1,:,:)
            p_cmask(   :,1   ,:) = p_cmask(     :,2   ,:)
            p_cmask(:,jpj_crs,:) = p_cmask(:,jpj_crsm1,:)
         ENDIF
      ENDIF

      IF ( PRESENT( p_cmask2d ) ) THEN
         IF( nperio /= 0 ) THEN
            CALL crs_lbc_lnk( cd_type,1.0,pt2d=p_cmask2d )
         ELSE
            p_cmask2d( 1  ,:   ) = p_cmask2d(     2,:   )          ! all points
            p_cmask2d(jpi_crs,:) = p_cmask2d(jpi_crsm1,:)
            p_cmask2d(   :,1   ) = p_cmask2d(     :,2   )
            p_cmask2d(:,jpj_crs) = p_cmask2d(:,jpj_crsm1)
         ENDIF
      ENDIF



   END SUBROUTINE crsfun_mask

   SUBROUTINE crsfun_coordinates( p_pgphi, p_pglam, cd_type, p_cgphi, p_cglam )
      !!----------------------------------------------------------------
      !!               *** SUBROUTINE crsfun_coordinates ***
      !! ** Purpose :  Determine the coordinates for the coarse grid
      !! 
      !! ** Method  :  From the parent grid subset, search for the central
      !!               point.  For an odd-numbered reduction factor,
      !!               the coordinate will be that of the central T-cell.
      !!               For an even-numbered reduction factor, of a non-square
      !!               coarse grid box, the coordinate will be that of 
      !!               the east or north face or more likely.  For a square
      !!               coarse grid box, the coordinate will be that of
      !!               the central f-corner.
      !!
      !! ** Input   :  p_pgphi = parent grid gphi[t|u|v|f]
      !!               p_pglam = parent grid glam[t|u|v|f]
      !!               cd_type  = grid type (T,U,V,F)
      !! ** Output  :  p_cgphi = coarse grid gphi[t|u|v|f]
      !!               p_cglam = coarse grid glam[t|u|v|f]
      !!              
      !! History. 1 Jun.
      !!----------------------------------------------------------------
      !! Arguments
      REAL(wp), DIMENSION(jpi,jpj), INTENT(in)  :: p_pgphi  ! Parent grid latitude
      REAL(wp), DIMENSION(jpi,jpj), INTENT(in)  :: p_pglam  ! Parent grid longitude
      CHARACTER(len=1),             INTENT(in)  :: cd_type   ! grid type (T,U,V,F) 
      REAL(wp), ALLOCATABLE, DIMENSION(:,:), INTENT(out) :: p_cgphi  ! Coarse grid latitude
      REAL(wp), ALLOCATABLE, DIMENSION(:,:), INTENT(out) :: p_cglam  ! Coarse grid longitude

      !! Local variables
      INTEGER ::  ji, jj, jk                   ! dummy loop indices
      INTEGER :: ijie,ijis,ijje,ijjs,ijpk

      WRITE(numout,*) 'crsfun_coordinates. begin'

      !! Initialize output fields
      IF ( .NOT. ALLOCATED( p_cgphi ) ) THEN
        ALLOCATE( p_cgphi(jpi_crs,jpj_crs), p_cglam(jpi_crs,jpj_crs) )
      ENDIF
      p_cgphi(:,:) = 0.e0
      p_cglam(:,:) = 0.e0


      DO ji = 2, jpi_crsm1

         IF ( cd_type == 'T' .OR. cd_type == 'V' )  ijis = mis_crs(ji) + mxbinctr 
         IF ( cd_type == 'U' .OR. cd_type == 'F' )  ijis = mie_crs(ji)

         DO jj = 2, jpj_crsm1
     
            IF ( cd_type == 'T' .OR. cd_type == 'U' ) ijjs = mjs_crs(jj) + mybinctr                  
            IF ( cd_type == 'V' .OR. cd_type == 'F' ) ijjs = mje_crs(jj)

            p_cgphi(ji,jj) = p_pgphi(ijis,ijjs)
            p_cglam(ji,jj) = p_pglam(ijis,ijjs)
             
         ENDDO

      ENDDO

      WRITE(numout,*) 'crsfun_coordinates. completed set new coords'

! Retroactively add back the boundary halo cells.

      IF( nperio /= 0 ) THEN
      WRITE(numout,*) 'crsfun_coordinates. call crs_lbc_lnk'

         CALL crs_lbc_lnk( cd_type,1.0,p_cgphi )
         CALL crs_lbc_lnk( cd_type,1.0,p_cglam )
      ELSE
            p_cgphi(      1,:) = p_cgphi(2        ,:)          ! all points
            p_cgphi(jpi_crs,:) = p_cgphi(jpi_crsm1,:)
            p_cgphi(      :,1) = p_cgphi(        :,2)
            p_cgphi(:,jpj_crs) = p_cgphi(:,jpj_crsm1)

            p_cglam(      1,:) = p_cglam(        2,:)          ! all points
            p_cglam(jpi_crs,:) = p_cglam(jpi_crsm1,:)
            p_cglam(      :,1) = p_cglam(        :,2)
            p_cglam(:,jpj_crs) = p_cglam(:,jpj_crsm1)

      ENDIF

      WRITE(numout,*) 'crsfun_coordinates. fill out edges'

      ! Fill up jrow=1 which is zeroed out or not handled by lbc_lnk and lbc_nfd
      DO ji = 2, jpi_crsm1

         IF ( cd_type == 'T' .OR. cd_type == 'V' )  ijis = mis_crs(ji) + mxbinctr 
         IF ( cd_type == 'U' .OR. cd_type == 'F' )  ijis = mie_crs(ji)

         p_cgphi(ji,1) = p_pgphi(ijis,1)
         p_cglam(ji,1) = p_pglam(ijis,1)

      ENDDO

      ! Fill i=1, i=jpi at j=1
      p_cgphi(1,1) = p_cgphi(jpi_crsm1,1)
      p_cglam(1,1) = p_cglam(jpi_crsm1,1)
      p_cgphi(jpi_crs,1) = p_cgphi(2,1)
      p_cglam(jpi_crs,1) = p_cglam(2,1)
      ! Fill upper-right corner i=1, j=jpj_crs
     !cc IF ( nperio == 4 ) THEN
     !cc    p_cgphi(1,jpj_crs) = p_cgphi(jpi_crsm1,jpj_crs-2)
     !cc    p_cglam(1,jpj_crs) = p_cglam(jpi_crsm1,jpj_crs-2)
     !cc ELSEIF ( nperio == 6 ) THEN
     !cc    p_cgphi(1,jpj_crs) = p_cgphi(jpi_crs,jpj_crsm1)
     !cc    p_cglam(1,jpj_crs) = p_cglam(jpi_crs,jpj_crsm1)
     !cc ENDIF

      WRITE(numout,*) 'crsfun_coordinates. done'


   END SUBROUTINE crsfun_coordinates

   SUBROUTINE crsfun_wgt( cd_type, cd_op, p_pmask, p_e1, p_e2, p_fse3, &
      &                   p_cfield2d_1, p_cfield2d_2, p_cfield3d_1, p_cfield3d_2 )
      !!----------------------------------------------------------------
      !!               *** SUBROUTINE crsfun_wgt ***
      !! ** Purpose :  Three applications.
      !!               1) SUM. Get coarse grid horizontal scale factors and unmasked fraction
      !!               2) VOL. Get coarse grid box volumes
      !!               3) WGT. Weighting multiplier for volume-weighted and/or
      !!                       area-weighted averages.
      !!                       Weights (i.e. the denominator) calculated here 
      !!                       to avoid IF-tests and division.
      !! ** Method  :  1) SUM.  For grid types T,U,V,F (and W) the 2D scale factors of 
      !!               the coarse grid are the sum of the east or north faces of the 
      !!               parent grid subset comprising the coarse grid box.  
      !!               The fractions of masked:total surface (3D) on the east, 
      !!               north and top faces is, optionally, also output.
      !!               - Top face area sum
      !!                 Valid arguments: cd_type, cd_op='W', p_pmask, p_e1, p_e2 
      !!               - Top face ocean surface fraction
      !!                 Valid arguments: cd_type, cd_op='W', p_pmask, p_e1, p_e2       
      !!               - e1,e2 Scale factors
      !!                 Valid arguments: 
      !!               2) VOL.  For grid types W and T, the coarse grid box
      !!               volumes are output. Also optionally, the fraction of  
      !!               masked:total volume of the parent grid subset is output (i.e. facvol).
      !!               3) WGT. Based on the grid type, the denominator is pre-determined here to  
      !!               perform area- or volume- weighted averages, 
      !!               to avoid IF-tests and divisions.
      !! ** Inputs  : p_e1, p_e2  = parent grid e1 or e2 (t,u,v,f)
      !!              p_pmask     = parent grid mask (T,U,V,F) 
      !!              cd_type     = grid type (T,U,V,F) for scale factors; for velocities (U or V)
      !!              cd_op       = applied operation (SUM, VOL, WGT)
      !!              p_fse3      = (Optional) parent grid vertical level thickness (fse3u or fse3v)
      !! ** Outputs : p_cfield2d_1 = (Optional) 2D field on coarse grid
      !!              p_cfield2d_2 = (Optional) 2D field on coarse grid
      !!              p_cfield3d_1 = (Optional) 3D field on coarse grid
      !!              p_cfield3d_2 = (Optional) 3D field on coarse grid
      !!
      !! History.     4 Jun.  Write for WGT and scale factors only
      !!----------------------------------------------------------------
      !! 
      !!  Arguments
      CHARACTER(len=1),                 INTENT(in) :: cd_type  ! grid type U,V 
      CHARACTER(len=3),                 INTENT(in) :: cd_op    ! operation sum or average
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(in) :: p_pmask  ! Parent grid U,V mask
      REAL(wp), DIMENSION(jpi,jpj),     INTENT(in) :: p_e1     ! Parent grid U,V scale factors (e1)
      REAL(wp), DIMENSION(jpi,jpj),     INTENT(in) :: p_e2     ! Parent grid U,V scale factors (e2)

      REAL(wp), DIMENSION(:,:),   INTENT(out), ALLOCATABLE, OPTIONAL :: p_cfield2d_1 ! Coarse grid box 2D quantity
      REAL(wp), DIMENSION(:,:),   INTENT(out), ALLOCATABLE, OPTIONAL :: p_cfield2d_2 ! Coarse grid box 2D quantity
      REAL(wp), DIMENSION(:,:,:), INTENT(out), ALLOCATABLE, OPTIONAL :: p_cfield3d_1 ! Coarse grid box 3D quantity 
      REAL(wp), DIMENSION(:,:,:), INTENT(out), ALLOCATABLE, OPTIONAL :: p_cfield3d_2 ! Coarse grid box 3D quantity 

      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(in), OPTIONAL :: p_fse3     ! Parent grid vertical level thickness (fse3u, fse3v)

      !! Local variables
      INTEGER                                 ::  ji, jj, jk, jii, jjj     ! dummy loop indices
      INTEGER                                 :: ijie,ijis,ijje,ijjs,ijpk
      REAL(wp)                                :: zdAm                      ! masked face area
      REAL(wp), DIMENSION(:,:),   ALLOCATABLE :: ze1, ze2
      REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: ze3       
      REAL(wp), DIMENSION(:,:),   ALLOCATABLE :: zcfield2d_1, zcfield2d_2
      REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: zcfield3d_1, zcfield3d_2
  
      !!----------------------------------------------------------------  
      ! Initialize      

      ! Arrays, scalars initialization 
      ALLOCATE( ze1(jpi,jpj) , ze2(jpi,jpj) )
      ALLOCATE( ze3(jpi,jpj,jpk) )
      ALLOCATE( zcfield2d_1(jpi_crs,jpj_crs) , zcfield2d_2(jpi_crs,jpj_crs) )
      ALLOCATE( zcfield3d_1(jpi_crs,jpj_crs,jpk), zcfield3d_2(jpi_crs,jpj_crs,jpk) )

      ze1(:,:) = 1.0
      ze2(:,:) = 1.0
      ze3(:,:,:) = 1.0
      zcfield2d_1(:,:)  = 0.0
      zcfield2d_2(:,:)  = 0.0
      zcfield3d_1(:,:,:) = 0.0
      zcfield3d_2(:,:,:) = 0.0
    
      ijpk = jpk

     ! Control of arguments
      ze1(:,:) = p_e1(:,:)
      ze2(:,:) = p_e2(:,:)

      IF ( PRESENT(p_cfield2d_1) ) ALLOCATE( p_cfield2d_1(jpi_crs,jpj_crs) )
      IF ( PRESENT(p_cfield2d_2) ) ALLOCATE( p_cfield2d_2(jpi_crs,jpj_crs) ) 
      IF ( PRESENT(p_cfield3d_1) ) ALLOCATE( p_cfield3d_1(jpi_crs,jpj_crs,jpk) ) 
      IF ( PRESENT(p_cfield3d_2) ) ALLOCATE( p_cfield3d_2(jpi_crs,jpj_crs,jpk) ) 

      IF ( PRESENT(p_cfield2d_1) ) p_cfield2d_1(:,:) = 0.0
      IF ( PRESENT(p_cfield2d_2) ) p_cfield2d_2(:,:) = 0.0
      IF ( PRESENT(p_cfield3d_1) ) p_cfield3d_1(:,:,:) = 0.0
      IF ( PRESENT(p_cfield3d_2) ) p_cfield3d_2(:,:,:) = 0.0

      IF ( PRESENT(p_fse3) ) ze3(:,:,:) = p_fse3(:,:,:)


          DO jk = 1, ijpk    

             zcfield2d_1(:,:) = 0.0 ; zcfield2d_2(:,:) = 0.0        
             ! DO ji = 2, jpi_crsm1
             DO ji = 1, jpi_crs
                ijie = mie_crs(ji)
                ijis = mis_crs(ji)

             !   DO jj = 1, jpj_crsm1
                DO jj = 1, jpj_crs
                   ijje = mje_crs(jj) 
                   ijjs = mjs_crs(jj)                   

                   IF ( cd_op == 'SUM' ) THEN

                      IF ( cd_type == 'W' ) THEN

                         ! Surface area of top of grid cell, e1e2 
                         DO jii = ijis, ijie
                            DO jjj = ijjs, ijje
                               zcfield2d_1(ji,jj) = zcfield2d_1(ji,jj) + ( ze1(jii,jjj) * ze2(jii,jjj) )
                            ENDDO
                         ENDDO

                         ! Surface ocean fraction, e1e2 masked/e1e2 total
                         IF ( zcfield2d_1(ji,jj) /= 0 ) THEN
                            DO jii = ijis, ijie
                               DO jjj = ijjs, ijje
                                  zcfield3d_1(ji,jj,jk) = zcfield3d_1(ji,jj,jk) + (  ze1(jii,jjj)           &
                                     &                                             * ze2(jii,jjj)          &
                                     &                                             * p_pmask(jii,jjj,jk) )
                               ENDDO
                            ENDDO
                            zcfield3d_1(ji,jj,jk) = zcfield3d_1(ji,jj,jk) / zcfield2d_1(ji,jj)
                         ENDIF

                      ELSE
                         SELECT CASE ( cd_type )
                         
                         CASE ( 'T' )
                            IF ( nn_factx == 3 ) THEN
                               ! Calculate e1 scale factor or if present ze3, unmasked surface area
                               IF (jj == 1) THEN
                                  DO jii = ijis, ijie
                                  zcfield2d_1(ji,jj) = zcfield2d_1(ji,jj) +  ( ze1(jii,ijje) * ze3(jii,ijje-1,jk) ) 
                                  ENDDO
                               ELSE
                                  DO jii = ijis, ijie
                                     zcfield2d_1(ji,jj) = zcfield2d_1(ji,jj) +  ( ze1(jii,ijje-1) * ze3(jii,ijje-1,jk) ) 
                                  ENDDO
                               ENDIF
                            
                               ! Calculate e2 scale factor
                               DO jjj = ijjs, ijje
                                zcfield2d_2(ji,jj) = zcfield2d_2(ji,jj) +  ( ze2(ijie-1,jjj) * ze3(ijie-1,jjj,jk) )
                               ENDDO
                            ENDIF
                            
                         CASE ( 'U' )
                            IF ( nn_factx == 3 ) THEN
                               ! Calculate e1 scale factor or if present ze3, unmasked surface area
                               IF (jj == 1) THEN
                                  DO jii = ijis, ijie
                                  zcfield2d_1(ji,jj) = zcfield2d_1(ji,jj) +  ( ze1(jii+1,ijje) * ze3(jii+1,ijje-1,jk) ) 
                                  ENDDO
                               ELSE
                                  DO jii = ijis, ijie
                                     zcfield2d_1(ji,jj) = zcfield2d_1(ji,jj) +  ( ze1(jii+1,ijje-1) * ze3(jii+1,ijje-1,jk) ) 
                                  ENDDO
                               ENDIF                            
                               ! Calculate e2 scale factor
                               DO jjj = ijjs, ijje
                                zcfield2d_2(ji,jj) = zcfield2d_2(ji,jj) +  ( ze2(ijie,jjj) * ze3(ijie,jjj,jk) )
                               ENDDO
                            ENDIF
                            
                         CASE ( 'V' )
                            IF ( nn_factx == 3 ) THEN
                               ! Calculate e1 scale factor or if present ze3, unmasked surface area
                               DO jii = ijis, ijie
                                  zcfield2d_1(ji,jj) = zcfield2d_1(ji,jj) +  ( ze1(jii,ijje) * ze3(jii,ijje,jk) ) 
                               ENDDO
                            
                               ! Calculate e2 scale factor
                               DO jjj = ijjs, ijje
                                zcfield2d_2(ji,jj) = zcfield2d_2(ji,jj) +  ( ze2(ijie-1,jjj+1) * ze3(ijie-1,jjj+1,jk) )
                               ENDDO
                            ENDIF
                            
                            CASE ( 'F' )
                            IF ( nn_factx == 3 ) THEN
                               ! Calculate e1 scale factor or if present ze3, unmasked surface area
                               DO jii = ijis, ijie
                                  zcfield2d_1(ji,jj) = zcfield2d_1(ji,jj) +  ( ze1(jii+1,ijje) * ze3(jii+1,ijje,jk) ) 
                               ENDDO
                            
                               ! Calculate e2 scale factor
                               DO jjj = ijjs, ijje
                                zcfield2d_2(ji,jj) = zcfield2d_2(ji,jj) +  ( ze2(ijie,jjj+1) * ze3(ijie,jjj+1,jk) )
                               ENDDO
                            ENDIF
                         END SELECT
                         
                         
                         
                         IF ( PRESENT(p_cfield3d_1) ) THEN

                            IF ( cd_type == 'V' ) THEN
                               IF ( zcfield2d_1(ji,jj) /= 0 ) THEN                           ! e1 ocean length or e1e3 surface fraction
                                  DO jii = ijis, ijie
                                     zcfield3d_1(ji,jj,jk) = zcfield3d_1(ji,jj,jk) &
                                        &                   + ( ze1(jii,ijje) * ze3(jii,ijje,jk) * p_pmask(jii,ijje,jk) ) 
                                  ENDDO
                                  zcfield3d_1(ji,jj,jk) = zcfield3d_1(ji,jj,jk) / zcfield2d_1(ji,jj) 
                               ENDIF
                            ENDIF

                            IF ( cd_type == 'U' ) THEN
                               IF ( zcfield2d_2(ji,jj) /= 0 ) THEN                           ! e2 ocean length or e2e3 surface fraction
                                  DO jjj = ijjs, ijje
                                     zcfield3d_1(ji,jj,jk) = zcfield3d_1(ji,jj,jk) &
                                        &                   + ( ze2(ijie,jjj) * ze3(ijie,jjj,jk) * p_pmask(ijie,jjj,jk) )
                                  ENDDO
                                  zcfield3d_1(ji,jj,jk) = zcfield3d_1(ji,jj,jk) / zcfield2d_2(ji,jj) 
                               ENDIF
                            ENDIF

                         ENDIF

                      ENDIF

                   ENDIF
                   
                   IF ( cd_op == 'POS' ) THEN      !cc
                   
                      IF ( nn_factx == 3 .AND. nn_facty == 3) THEN
                      
                         SELECT CASE ( cd_type )
                         
                            CASE ( 'T' )
                            
                               IF ((jj == 1) .AND. (ji == 1)) THEN
                                  ! Calculate e1 scale factor or if present ze3, unmasked surface area
                                  zcfield2d_1(ji,jj) = ( ze1(ijie,ijje  ) * ze3(ijie,ijje,jk) ) * nn_factx
                            
                                  ! Calculate e2 scale factor
                                  zcfield2d_2(ji,jj) = ( ze2(ijie,ijje  ) * ze3(ijie,ijje,jk) ) * nn_facty
                               ELSEIF (jj == 1) THEN
                                  ! Calculate e1 scale factor or if present ze3, unmasked surface area
                                  zcfield2d_1(ji,jj) = ( ze1(ijie-1,ijje  ) * ze3(ijie-1,ijje,jk) ) * nn_factx
                            
                                  ! Calculate e2 scale factor
                                  zcfield2d_2(ji,jj) = ( ze2(ijie-1,ijje  ) * ze3(ijie-1,ijje,jk) ) * nn_facty
                               ELSEIF (ji == 1) THEN 
                                  ! Calculate e1 scale factor or if present ze3, unmasked surface area
                                  zcfield2d_1(ji,jj) = ( ze1(ijie,ijje-1) * ze3(ijie,ijje-1,jk) ) * nn_factx
                            
                                  ! Calculate e2 scale factor
                                  zcfield2d_2(ji,jj) = ( ze2(ijie,ijje-1) * ze3(ijie,ijje-1,jk) ) * nn_facty
                               ELSE
                                 ! Calculate e1 scale factor or if present ze3, unmasked surface area
                                  zcfield2d_1(ji,jj) = ( ze1(ijie-1,ijje-1) * ze3(ijie-1,ijje-1,jk) ) * nn_factx
                            
                                  ! Calculate e2 scale factor
                                  zcfield2d_2(ji,jj) = ( ze2(ijie-1,ijje-1) * ze3(ijie-1,ijje-1,jk) ) * nn_facty
                               ENDIF
                               
                            CASE ( 'U' )
                               IF (jj == 1) THEN
                                  ! Calculate e1 scale factor or if present ze3, unmasked surface area
                                  zcfield2d_1(ji,jj) = ( ze1(ijie  ,ijje  ) * ze3(ijie  ,ijje,jk) ) * nn_factx
                            
                                  ! Calculate e2 scale factor
                                  zcfield2d_2(ji,jj) = ( ze2(ijie  ,ijje  ) * ze3(ijie  ,ijje,jk) ) * nn_facty
                               ELSE
                                  ! Calculate e1 scale factor or if present ze3, unmasked surface area
                                  zcfield2d_1(ji,jj) = ( ze1(ijie  ,ijje-1) * ze3(ijie  ,ijje-1,jk) ) * nn_factx
                            
                                  ! Calculate e2 scale factor
                                  zcfield2d_2(ji,jj) = ( ze2(ijie  ,ijje-1) * ze3(ijie  ,ijje-1,jk) ) * nn_facty
                              ENDIF
                                
                            CASE ( 'V' )
                               IF (ji == 1) THEN 
                               ! Calculate e1 scale factor or if present ze3, unmasked surface area
                               zcfield2d_1(ji,jj) = ( ze1(ijie,ijje  ) * ze3(ijie,ijje  ,jk) ) * nn_factx
                            
                               ! Calculate e2 scale factor
                               zcfield2d_2(ji,jj) = ( ze2(ijie,ijje  ) * ze3(ijie,ijje  ,jk) ) * nn_facty
                               ELSE
                               ! Calculate e1 scale factor or if present ze3, unmasked surface area
                               zcfield2d_1(ji,jj) = ( ze1(ijie-1,ijje  ) * ze3(ijie-1,ijje  ,jk) ) * nn_factx
                            
                               ! Calculate e2 scale factor
                               zcfield2d_2(ji,jj) = ( ze2(ijie-1,ijje  ) * ze3(ijie-1,ijje  ,jk) ) * nn_facty
                               ENDIF
                               
                            CASE ( 'F' )
                               ! Calculate e1 scale factor or if present ze3, unmasked surface area
                               zcfield2d_1(ji,jj) = ( ze1(ijie  ,ijje  ) * ze3(ijie  ,ijje  ,jk) ) * nn_factx
                            
                               ! Calculate e2 scale factor
                               zcfield2d_2(ji,jj) = ( ze2(ijie  ,ijje  ) * ze3(ijie  ,ijje  ,jk) ) * nn_facty
                               
                         END SELECT
                      ENDIF
                   ENDIF                    !cc
                   

                   IF ( cd_op == 'WGT' ) THEN

                      zdAm = 0.0

                      IF ( cd_type == 'V' ) THEN
                         ! 
                         DO jii = ijis, ijie
                            zdAm = zdAm + ( ze1(jii,ijje) * ze3(jii,ijje,jk) * p_pmask(jii,ijje,jk) ) 
                         ENDDO
                         IF ( zdAm /= 0 ) zcfield3d_1(ji,jj,jk) = zdAm 
    
                      ELSEIF ( cd_type == 'U') THEN
                         DO jjj = ijjs, ijje
                            zdAm = zdAm + ( ze2(ijie,jjj) * ze3(ijie,jjj,jk) * p_pmask(ijie,jjj,jk) )
                         ENDDO
                         IF ( zdAm /= 0 ) zcfield3d_1(ji,jj,jk) = zdAm 

                      ELSEIF ( cd_type == 'W' ) THEN
                         DO jii = ijis, ijie
                            DO jjj = ijjs, ijje
                               zdAm = zdAm + ( ze1(jii,jjj) * ze2(jii,jjj) * p_pmask(jii,jjj,jk) )
                            ENDDO
                         ENDDO
                         IF ( zdAm /= 0 ) zcfield3d_1(ji,jj,jk) = zdAm

                      ELSEIF ( cd_type == 'T' ) THEN
                         DO jii = ijis, ijie
                            DO jjj = ijjs, ijje
                               zdAm = zdAm + ( ze1(jii,jjj) * ze2(jii,jjj) * ze3(jii,jjj,jk) * p_pmask(jii,jjj,jk) )
                            ENDDO
                         ENDDO
                         IF ( zdAm /= 0 ) zcfield3d_1(ji,jj,jk) = zdAm

                      ELSE
  
                         ! jes. Add a stop?

                      ENDIF

                   ENDIF

                   IF ( cd_op == 'VOL' ) THEN

                      zdAm = 0.0

                      IF ( cd_type == 'W' .OR. cd_type == 'T' ) THEN

                         DO jii = ijis, ijie
                            DO jjj = ijjs, ijje
                               zcfield3d_1(ji,jj,jk) = zcfield3d_1(ji,jj,jk) + ( ze1(jii,jjj) * ze2(jii,jjj) * ze3(jii,jjj,jk) )
                               zdAm = zdAm + ( ze1(jii,jjj) * ze2(jii,jjj) * ze3(jii,jjj,jk) * p_pmask(jii,jjj,jk) )
                            ENDDO
                         ENDDO
                         IF ( zcfield3d_1(ji,jj,jk) /= 0 ) zcfield3d_2(ji,jj,jk) = zdAm / zcfield3d_1(ji,jj,jk)

                      ELSE
                          ! jes. add a stop?
                      ENDIF

                   ENDIF
                                    
                ENDDO
             ENDDO

          ENDDO

! Retroactively add back the boundary halo cells.

      IF( nperio /= 0 ) THEN

         ! Take care of the 2D arrays
         IF ( cd_op == 'SUM' .OR.  cd_op == 'POS') THEN
            IF ( PRESENT(p_cfield2d_1) ) THEN
               p_cfield2d_1(:,:) = zcfield2d_1(:,:)
               CALL crs_lbc_lnk( cd_type,1.0,pt2d=p_cfield2d_1 )

               ! Fill up jrow=1 which is zeroed out or not handled by lbc_lnk and lbc_nfd
               p_cfield2d_1(:,1) = zcfield2d_1(:,1)  !cc 
               ! Fill i=1, i=jpi at j=1
               p_cfield2d_1(1,1) = p_cfield2d_1(jpi_crsm1,1)
               p_cfield2d_1(jpi_crs,1) = p_cfield2d_1(2,1)
               
             !cc  p_cfield2d_1(1,jpj_crs-1) = p_cfield2d_1(3,jpj_crs-1)

              ! Fill upper-right corner i=1, j=jpj_crs
              !cc IF ( nperio == 4 ) THEN on écrase les valeurs limites calculées dans crs_lbc_lnk
              !cc   p_cfield2d_1(1,jpj_crs) = p_cfield2d_1(jpi_crsm1,jpj_crs-2)
              !cc ELSEIF ( nperio == 6 ) THEN
              !cc    p_cfield2d_1(1,jpj_crs) = p_cfield2d_1(jpi_crs,jpj_crsm1)
              !cc ENDIF

            ENDIF
            IF ( PRESENT(p_cfield2d_2) ) THEN
               p_cfield2d_2(:,:) = zcfield2d_2(:,:)
               CALL crs_lbc_lnk( cd_type,1.0,pt2d=p_cfield2d_2 )
               
               ! Fill up jrow=1 which is zeroed out or not handled by lbc_lnk and lbc_nfd
               p_cfield2d_2(:,1) = zcfield2d_2(:,1)

               ! Fill i=1, i=jpi at j=1
               p_cfield2d_2(1,1) = p_cfield2d_2(jpi_crsm1,1)
               p_cfield2d_2(jpi_crs,1) = p_cfield2d_2(2,1)
               IF ( cd_op == 'SUM') THEN 
                  DO jii = 1 , jpiglo_crs
                  p_cfield2d_2(jii,1) = p_cfield2d_2(jii,1) * 3
                  ENDDO
               ENDIF
               ! Fill upper-right corner i=1, j=jpj_crs
           !cc     IF ( nperio == 4 ) THEN
           !cc       p_cfield2d_2(1,jpj_crs) = p_cfield2d_2(jpi_crsm1,jpj_crs-2)
           !cc    ELSEIF ( nperio == 6 ) THEN
           !cc       p_cfield2d_2(1,jpj_crs) = p_cfield2d_2(jpi_crs,jpj_crsm1)
           !cc ENDIF
            ENDIF

         ELSE

            CALL crs_lbc_lnk( cd_type,1.0,pt2d=zcfield2d_1 ) 
            IF ( PRESENT(p_cfield2d_1) ) p_cfield2d_1(:,:) = zcfield2d_1(:,:)
            CALL crs_lbc_lnk( cd_type,1.0,pt2d=zcfield2d_2 ) 
            IF ( PRESENT(p_cfield2d_2) ) p_cfield2d_2(:,:) = zcfield2d_2(:,:)

         ENDIF

         ! Take care now of 3d arrays
         IF ( cd_op == 'SUM' .OR. cd_op == 'VOL' .OR. cd_op == 'POS'  ) THEN
            CALL crs_lbc_lnk( cd_type,1.0,pt3d1=zcfield3d_1 ) 
            IF ( PRESENT(p_cfield3d_1) ) p_cfield3d_1(:,:,:) = zcfield3d_1(:,:,:)
            CALL crs_lbc_lnk( cd_type,1.0,pt3d1=zcfield3d_2 ) 
            IF ( PRESENT(p_cfield3d_2) ) p_cfield3d_2(:,:,:) = zcfield3d_2(:,:,:)
         ELSE
            p_cfield3d_1(:,:,:) = zcfield3d_1(:,:,:)
            CALL crs_lbc_lnk( cd_type,1.0,pt3d1=p_cfield3d_1 )

            ! Fill upper-right corner i=1, j=jpj_crs
            IF ( nperio == 4 ) THEN
               p_cfield3d_1(1,jpj_crs,:) = p_cfield3d_1(jpi_crsm1,jpj_crs-2,:)
            ELSEIF ( nperio == 6 ) THEN
               p_cfield3d_1(1,jpj_crs,:) = p_cfield3d_1(jpi_crs,jpj_crsm1,:)
            ENDIF

         ENDIF

      ELSE
         IF ( cd_op == 'SUM' .OR. cd_op == 'POS'  ) THEN      
            IF ( PRESENT(p_cfield2d_1) ) THEN
               p_cfield2d_1(:,:) = zcfield2d_1(:,:)
               p_cfield2d_1(      1,:) = p_cfield2d_1(        2,:)          ! all points
               p_cfield2d_1(jpi_crs,:) = p_cfield2d_1(jpi_crsm1,:)
               p_cfield2d_1(      :,1) = p_cfield2d_1(        :,2)
               p_cfield2d_1(:,jpj_crs) = p_cfield2d_1(:,jpj_crsm1)
            ENDIF
            IF ( PRESENT(p_cfield2d_2) ) THEN
               p_cfield2d_2(:,:) = zcfield2d_2(:,:)
               p_cfield2d_2(      1,:) = p_cfield2d_2(        2,:)          ! all points
               p_cfield2d_2(jpi_crs,:) = p_cfield2d_2(jpi_crsm1,:)
               p_cfield2d_2(      :,1) = p_cfield2d_2(        :,2)
               p_cfield2d_2(:,jpj_crs) = p_cfield2d_2(:,jpj_crsm1)
            ENDIF
         ENDIF 

         IF ( PRESENT( p_cfield3d_1 ) ) THEN
            p_cfield3d_1(:,:,:) = zcfield3d_1(:,:,:)
            p_cfield3d_1( 1     ,:,:) = p_cfield3d_1(   2     ,:,:)      ! all points
            p_cfield3d_1(jpi_crs,:,:) = p_cfield3d_1(jpi_crsm1,:,:)
            p_cfield3d_1(   :   ,1,:) = p_cfield3d_1(   :     ,2,:)
            p_cfield3d_1(:,jpj_crs,:) = p_cfield3d_1(:,jpj_crsm1,:)
         ENDIF
         IF ( PRESENT( p_cfield3d_2 ) ) THEN
            p_cfield3d_2(:,:,:) = zcfield3d_2(:,:,:)
            p_cfield3d_2( 1     ,:,:) = p_cfield3d_2(   2     ,:,:)      ! all points
            p_cfield3d_2(jpi_crs,:,:) = p_cfield3d_2(jpi_crsm1,:,:)
            p_cfield3d_2(   :   ,1,:) = p_cfield3d_2(   :     ,2,:)
            p_cfield3d_2(:,jpj_crs,:) = p_cfield3d_2(:,jpj_crsm1,:)
         ENDIF
      ENDIF

      DEALLOCATE( ze1 , ze2 )
      DEALLOCATE( ze3 )
      DEALLOCATE( zcfield2d_1 , zcfield2d_2 )
      DEALLOCATE( zcfield3d_1 , zcfield3d_2 )

   END SUBROUTINE crsfun_wgt


   SUBROUTINE crsfun_UV( p_e1_e2, cd_type, psgn, p_pmask, p_fse3, p_pfield, p_surf_crs, p_cfield3d )
      !!----------------------------------------------------------------
      !!               *** SUBROUTINE crsfun_UV ***
      !! ** Purpose :  Average, area-weighted, of U or V on the east and north faces 
      !!
      !! ** Method  :  The U and V velocities (3D) are determined as the area-weighted averages
      !!               on the east and north faces, respectively,
      !!               of the parent grid subset comprising the coarse grid box. 
      !!               In the case of the V and F grid, the last jrow minus 1 is spurious.
      !! ** Inputs  : p_e1_e2     = parent grid e1 or e2 (t,u,v,f)
      !!              cd_type     = grid type (T,U,V,F) for scale factors; for velocities (U or V)
      !!              psgn        = sign change over north fold (See lbclnk.F90)
      !!              p_pmask     = parent grid mask (T,U,V,F) for scale factors; 
      !!                                       for velocities (U or V)
      !!              p_fse3      = parent grid vertical level thickness (fse3u or fse3v)
      !!              p_pfield    = U or V on the parent grid
      !!              p_surf_crs  = (Optional) Coarse grid weight for averaging
      !! ** Outputs : p_cfield3d = 3D field on coarse grid
      !!
      !! History.  29 May.  completed draft.
      !!            4 Jun.  Revision for WGT
      !!            5 Jun.  Streamline for area-weighted average only ; separate scale factor and weights.
      !!----------------------------------------------------------------
      !! 
      !!  Arguments
      REAL(wp), DIMENSION(jpi,jpj),     INTENT(in)     :: p_e1_e2    ! Parent grid U,V scale factors (e1 or e2)
      CHARACTER(len=1),                 INTENT(in)     :: cd_type    ! grid type U,V 
      REAL(wp),                         INTENT(in)     :: psgn       ! sign change option across north fold
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(in)     :: p_pmask    ! Parent grid U,V mask
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(in)     :: p_fse3     ! Parent grid vertical level thickness (fse3u, fse3v)
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(in)     :: p_pfield   ! U or V on parent grid
      REAL(wp), DIMENSION(jpi_crs,jpj_crs,jpk), INTENT(in), OPTIONAL :: p_surf_crs ! Coarse grid area-weighting denominator    

      REAL(wp), DIMENSION(jpi_crs,jpj_crs,jpk), INTENT(out)  :: p_cfield3d ! Coarse grid box 3D quantity 

      !! Local variables
      INTEGER  :: ji, jj, jk , jii, jjj                  ! dummy loop indices
      INTEGER  :: ijie, ijis, ijje, ijjs
      REAL(wp), DIMENSION(:,:,:), ALLOCATABLE :: zsurfcrs   

      !!----------------------------------------------------------------  

      ALLOCATE( zsurfcrs(jpi_crs,jpj_crs,jpk) )
      zsurfcrs(:,:,:) = 1.0
      IF ( PRESENT(p_surf_crs) ) THEN
         WHERE ( p_surf_crs /= 0 ) zsurfcrs(:,:,:) = 1.0/p_surf_crs(:,:,:)
      ENDIF

      DO jk = 1, jpk    

         DO ji = 2, jpi_crsm1
            ijie = mie_crs(ji)
            ijis = mis_crs(ji)

            DO jj = 2, jpj_crsm1
               ijje = mje_crs(jj) 
               ijjs = mjs_crs(jj)                   

               IF ( cd_type == 'V' ) THEN

                  DO jii = ijis, ijie
                     p_cfield3d(ji,jj,jk) = p_cfield3d(ji,jj,jk) + ( p_pfield(jii,ijje,jk) * p_e1_e2(jii,ijje) * p_fse3(jii,ijje,jk) * p_pmask(jii,ijje,jk) ) 
                  ENDDO
                  p_cfield3d(ji,jj,jk) = p_cfield3d(ji,jj,jk) * zsurfcrs(ji,jj,jk) 

               ELSEIF ( cd_type == 'U') THEN

                  DO jjj = ijjs, ijje
                     p_cfield3d(ji,jj,jk) = p_cfield3d(ji,jj,jk) + ( p_pfield(ijie,jjj,jk) * p_e1_e2(ijie,jjj) * p_fse3(ijie,jjj,jk) * p_pmask(ijie,jjj,jk) )
                  ENDDO
                  p_cfield3d(ji,jj,jk) = p_cfield3d(ji,jj,jk) * zsurfcrs(ji,jj,jk)

               ENDIF
                 
            ENDDO
         ENDDO
      ENDDO

! Retroactively add back the boundary halo cells.

      IF( nperio /= 0 ) THEN
         CALL crs_lbc_lnk( cd_type,psgn,pt3d1=p_cfield3d )
      ELSE
         p_cfield3d( 1     ,:,:) = p_cfield3d(   2     ,:,:)      ! all points
         p_cfield3d(jpi_crs,:,:) = p_cfield3d(jpi_crsm1,:,:)
         p_cfield3d(   :   ,1,:) = p_cfield3d(   :     ,2,:)
         p_cfield3d(:,jpj_crs,:) = p_cfield3d(:,jpj_crsm1,:)
      ENDIF

      DEALLOCATE( zsurfcrs )

   END SUBROUTINE crsfun_UV

   SUBROUTINE crsfun_TW( p_e1e2t, cd_type, cd_op, p_cmask, p_ptmask, psgn, p_pfield2d, p_pfield3d_1, p_pfield3d_2, &
      &                  p_cfield2d, p_cfield3d)
      !!----------------------------------------------------------------
      !!               *** SUBROUTINE crsfun_TW ***
      !! ** Purpose :  Five applications.
      !!               1) Maximum surface quantity 
      !!                  - Vertical scale factors (fse3t or fse3w) 
      !!                    max thickness of the parent grid for coarse grid scale factors.
      !!                  - or diffusion test
      !!               2) Area-weighted mean quantity: w, or other 3D or 2D quantity
      !!               3) Volume-weighted mean quantity: tracer
      !!               4) Minimum surface quantity (diffusion test)
      !!               5) Area- or Volume- weighted sum.
      !! ** Method  :  1) - cd_op = 'MAX'. Determines the max vertical thickness of grid boxes
      !!                    including partial steps for at the bottom
      !!                    for the coarsened grid, where within the subset of 
      !!                    the parent grid cells the maximum thickness is taken.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!                    Where, normally p_pfield3d_1 is e3t.
      !!                  - cd_op = 'MAX'. May also be used for say, determining the maximum value of Kz, 
      !!                    thus p_pfield3d_1 is set to the 3D field, Kz.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!               2) - cd_op = 'ARE'. Calculate the area-weighted average (surface e1t*e2t)  
      !!                    of vertical velocity, w.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!                  - cd_op = 'ARE'. Calculate area-weighted average of a 2D quantity (e.g. emp)
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield2d
      !!               3) - cd_op = 'VOL'. Calculate the ocean volume (e1e2t*[fse3t|fse3w]) 
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!                  - cd_op = 'VOL'. Calculate volume-weighted average (volume e1t*e2t*fse3t) of a quantity.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1, p_pfield3d_2
      !!               4) - cd_op = 'MIN'. Calculate the minimum value on surface e1t*e2t for 3D variables
      !!                  Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!               5) - cd_op = 'SUM'. Calculate a dimesionally-weighted sum.  This could be area-weighted
      !!                  or volume-weighted sum. 
      !! ** Inputs  : p_e1e2t      = parent grid top face surface area, e1t*e2t 
      !!              cd_type      = grid type T, W (U, V, F) 
      !!              cd_op        = MAX, ARE, VOL, MIN, SUM
      !!              p_cmask      =  coarse grid mask
      !!              p_ptmask     =  parent grid tmask     
      !!              psgn         = (Optional) sign for lbc_lnk  
      !!              p_pfield2d   = (Optional) 2D field on parent grid
      !!              p_pfield3d_1 = (Optional) parent grid fse3t or fse3w
      !!              p_pfield3d_2 = (Optional) 3D field on parent grid
      !! ** Outputs : p_cfield2d   = (Optional) 2D field on coarse grid
      !!              p_cfield3d   = (Optional) 3D field on coarse grid
      !!
      !! 
      !! History.  30 May.  Editing.  To decide later: Keep all functionality or separate out the mean function.
      !!            7 Jun   TODO. Need to fix up the parent grid mask to optional like crsfun_UV...
      !!----------------------------------------------------------------
      !! 
      !!  Arguments
      REAL(wp), DIMENSION(jpi,jpj),               INTENT(in) :: p_e1e2t      ! Parent grid T surface (e1*e2)
      CHARACTER(len=1),                           INTENT(in) :: cd_type      ! grid type T, W ( U, V, F)
      CHARACTER(len=3),                           INTENT(in) :: cd_op        ! operation max, min, area-average, volume-average
      REAL(wp), DIMENSION(jpi_crs,jpj_crs,jpk),   INTENT(in) :: p_cmask      ! Coarse grid T mask
      REAL(wp), DIMENSION(jpi,jpj,jpk),           INTENT(in) :: p_ptmask     ! Parent grid T mask
      REAL(wp),                         OPTIONAL, INTENT(in) :: psgn         ! sign for lbc_lnk
      REAL(wp), DIMENSION(jpi,jpj),     OPTIONAL, INTENT(in) :: p_pfield2d   ! 2D quantity on parent grid, e.g. ssh
      REAL(wp), DIMENSION(jpi,jpj,jpk), OPTIONAL, INTENT(in) :: p_pfield3d_1 ! Normally parent grid vertical level thickness
      REAL(wp), DIMENSION(jpi,jpj,jpk), OPTIONAL, INTENT(in) :: p_pfield3d_2 ! 3D tracer or W on parent grid

      REAL(wp), DIMENSION(jpi_crs,jpj_crs),     OPTIONAL, INTENT(out):: p_cfield2d ! Coarse grid box east or north face quantity
      REAL(wp), DIMENSION(jpi_crs,jpj_crs,jpk), OPTIONAL, INTENT(out):: p_cfield3d ! Coarse grid box east or north face quantity 

      !! Local variables
      INTEGER ::  ji, jj, jk                   ! dummy loop indices
      INTEGER :: ijie,ijis,ijje,ijjs,ijpk,jii,jjj
      INTEGER, DIMENSION(3) :: idims
      REAL(wp), ALLOCATABLE, DIMENSION(:,:)   :: ze1e2, zpfield2d, zcfield2d
      REAL(wp), ALLOCATABLE, DIMENSION(:,:,:) :: ze3, zpfield3d, zcfield3d, zcmask, zpmask  
      REAL(wp)                                :: zdAm, zsgn
      !!----------------------------------------------------------------  
      ! Initialize

      ALLOCATE( ze3(jpi,jpj,jpk) , zpfield3d(jpi,jpj,jpk) , zpmask(jpi,jpj,jpk) )
      ALLOCATE( ze1e2(jpi,jpj) , zpfield2d(jpi,jpj) )
      ALLOCATE( zcfield2d(jpi_crs,jpj_crs) )
      ALLOCATE( zcfield3d(jpi_crs,jpj_crs,jpk) , zcmask(jpi_crs,jpj_crs,jpk) )

      ! Arrays, scalars initialization 
      zpfield2d(:,:)   = 0.0
      zpfield3d(:,:,:) = 0.0
      zcfield2d(:,:)   = 0.0
      zcfield3d(:,:,:) = 0.0
      zpmask(:,:,:)    = 1.0
      idims(:)         = 1

      zcmask(:,:,:)    = p_cmask(:,:,:)     
      zpmask(:,:,:)    = p_ptmask(:,:,:)

      ijpk             = jpk


      ! Control of optional arguments
      !
      IF ( PRESENT(psgn) ) THEN
         zsgn = psgn
      ELSE
         zsgn = 1.0 
      ENDIF
      !
      IF ( TRIM(cd_op) == 'MAX' ) THEN
         ! Find the maximum thickness in each parent grid subset
         IF ( PRESENT(p_pfield3d_1) )  THEN
            zpfield3d(:,:,:) = p_pfield3d_1(:,:,:)
            ze3(:,:,:) = 0.0
            ze1e2(:,:) = 0.0
         ELSE 
            WRITE(numout,*) 'crsfun_TW. MAX only 3D arrays supported' 
         ENDIF 
      ELSEIF ( TRIM(cd_op) == 'VOL' ) THEN
         IF ( PRESENT(p_pfield3d_1) ) THEN
            ze3(:,:,:)       = p_pfield3d_1(:,:,:)
            IF ( PRESENT(p_pfield3d_2) ) THEN
!              ! Prep to calculate a volume-averaged mean
               zpfield3d(:,:,:) = p_pfield3d_2(:,:,:)
               ze1e2(:,:)       = p_e1e2t(:,:)     
            ELSE
               WRITE(numout,*) 'crsfun_TW. WARNING. Supply both e3t and the field for volume-averaged field'
            ENDIF
         ELSE
            WRITE(numout,*) 'crsfun_TW. VOL only 3D arrays supported'  
         ENDIF
      ELSEIF ( TRIM(cd_op) == 'ARE' ) THEN
         ze1e2(:,:) = p_e1e2t(:,:)
         ze3(:,:,:) = 1.0
         IF ( PRESENT(p_pfield3d_1) ) THEN
            ! Prep to do the area-weighted average of (3D) W
            zpfield3d(:,:,:) = p_pfield3d_1(:,:,:)
         ELSEIF ( PRESENT(p_pfield2d) ) THEN
            ! Prep to do the area-weighted average of some 2D quantity 
            zpfield2d(:,:) = p_pfield2d(:,:) 
            ijpk=1
         ENDIF
      ELSEIF ( TRIM(cd_op) == 'MIN' ) THEN
         IF ( PRESENT(p_pfield3d_1) ) THEN
            ! Prep to do get the minimum diffusion on the top face
            zpfield3d(:,:,:) = p_pfield3d_1(:,:,:)
            ze3(:,:,:) = 0.0
            ze1e2(:,:) = 0.0
         ELSE
            WRITE(numout,*) 'crsfun_TW. MIN Only 3D arrays supported' 
         ENDIF
      ELSEIF ( TRIM(cd_op) == 'SUM' ) THEN
         ze1e2(:,:) = p_e1e2t(:,:)
         zpmask(:,:,:) = p_ptmask(:,:,:)
         ze3(:,:,:) = 1.0
        IF ( PRESENT(p_pfield3d_1) ) THEN
            IF ( PRESENT(p_pfield3d_2) ) THEN
!              ! Prep to calculate a volume-weighted sum
               zpfield3d(:,:,:) = p_pfield3d_2(:,:,:)
               ze3(:,:,:) = p_pfield3d_1(:,:,:)
            ELSE
               ! Prep to do the area-weighted sum of (3D) W
               zpfield3d(:,:,:) = p_pfield3d_1(:,:,:)
            ENDIF
         ELSEIF ( PRESENT(p_pfield2d) ) THEN
            ! Prep to do the area-weighted sum of some 2D quantity 
            zpfield2d(:,:) = p_pfield2d(:,:) 
            ijpk=1
         ENDIF 
      ELSE
         WRITE(numout,*) 'crsfun_TW. valid cd_op are MAX, MIN, ARE, VOL, SUM' 
      ENDIF

      ! Determine output 
      DO jk = 1, ijpk    

         IF ( ijpk == jpk ) zpfield2d(:,:) = zpfield3d(:,:,jk) 
         zcfield2d(:,:) = 0.0 

            DO ji = 2, jpi_crsm1
               ijie = mie_crs(ji)
               ijis = mis_crs(ji)

            ! DO jj = 2, jpj_crsm1
              DO jj = 1, jpj_crsm1
                  ijje = mje_crs(jj) 
                  ijjs = mjs_crs(jj)                   

                  ! First determine weighted sums 
                  IF ( TRIM(cd_op) == 'SUM' .OR. TRIM(cd_op) == 'ARE' .OR. TRIM(cd_op) == 'VOL' ) THEN
                     ! Area- or volume- weighted sum
                     ! Accumulate to sum in parent grid subset 
                    DO jii = ijis, ijie
                        DO jjj = ijjs, ijje
                           zcfield2d(ji,jj) = zcfield2d(ji,jj) &
                              &              + ( zpfield2d(jii,jjj)      &
                              &                  *   ze1e2(jii,jjj)      &
                              &                  *     ze3(jii,jjj,jk)   &
                              &                  *  zpmask(jii,jjj,jk) ) 

                        ENDDO
                     ENDDO

                  ENDIF

                  ! Calculate weighted average if desired
                  IF ( TRIM(cd_op) == 'ARE' .OR. TRIM(cd_op) == 'VOL' ) THEN

                     ! Area- or volume- weighted mean
                     ! Sum first parent grid subset  
                     zdAm = 0.0
                     DO jii = ijis, ijie
                        DO jjj = ijjs, ijje
                           zdAm = zdAm + (   ze1e2(jii,jjj)               &
                              &           *    ze3(jii,jjj,jk)            &
                              &           * zpmask(jii,jjj,jk) )
                        ENDDO
                     ENDDO
                     
                     IF ( zdAm /= 0 )  zcfield2d(ji,jj) = zcfield2d(ji,jj) / zdAm

                  ENDIF


                  IF ( TRIM(cd_op) == 'MAX' ) THEN
                     ! Find max in parent grid subset 
                     DO jii = ijis, ijie
                        DO jjj = ijjs, ijje
                           zcfield2d(ji,jj) = MAX( zcfield2d(ji,jj), zpfield3d(jii,jjj,jk)*zpmask(jii,jjj,jk) ) 
                        ENDDO
                     ENDDO
                  ENDIF

                  IF ( TRIM(cd_op) == 'MIN' ) THEN
                     ! Find min in parent grid subset  
                     DO jii = ijis, ijie
                        DO jjj = ijjs, ijje
                           IF ( zpmask(jii,jjj,jk) == 1 ) THEN
                              zcfield2d(ji,jj) = MIN( zcfield2d(ji,jj), zpfield3d(jii,jjj,jk) ) 
                           ENDIF
                        ENDDO
                     ENDDO
                  ENDIF

               ENDDO
            ENDDO


            IF ( ijpk == 1 ) THEN
               IF ( PRESENT(p_cfield2d) ) p_cfield2d(:,:) = zcfield2d(:,:) * zcmask(:,:,jk)
            ELSE
               IF ( PRESENT(p_cfield3d) ) p_cfield3d(:,:,jk) = zcfield2d(:,:) * zcmask(:,:,jk)
            ENDIF    

      ENDDO


! Retroactively add back the boundary halo cells.

      IF( nperio /= 0 ) THEN
         IF ( ijpk == 1 ) THEN
            IF ( PRESENT(p_cfield2d) ) CALL crs_lbc_lnk( cd_type,zsgn,pt2d=p_cfield2d )
         ELSE
            IF ( PRESENT(p_cfield3d) ) THEN
               CALL crs_lbc_lnk( cd_type,zsgn,pt3d1=p_cfield3d )
            ENDIF
         ENDIF
      ELSE
         IF ( ijpk == 1 ) THEN 
            IF ( PRESENT(p_cfield2d) ) THEN     
               p_cfield2d(      1,:) = p_cfield2d(        2,:)          ! all points
               p_cfield2d(jpi_crs,:) = p_cfield2d(jpi_crsm1,:)
               p_cfield2d(      :,1) = p_cfield2d(        :,2)
               p_cfield2d(:,jpj_crs) = p_cfield2d(:,jpj_crsm1)
            ENDIF
         ELSE
            IF ( PRESENT(p_cfield3d) ) THEN
               p_cfield3d( 1     ,:,:) = p_cfield3d(   2     ,:,:)      ! all points
               p_cfield3d(jpi_crs,:,:) = p_cfield3d(jpi_crsm1,:,:)
               p_cfield3d(   :   ,1,:) = p_cfield3d(   :     ,2,:)
               p_cfield3d(:,jpj_crs,:) = p_cfield3d(:,jpj_crsm1,:)
            ENDIF
         ENDIF
      ENDIF

      DEALLOCATE( ze3 , zpfield3d , zpmask )
      DEALLOCATE( ze1e2 , zpfield2d )
      DEALLOCATE( zcfield2d )
      DEALLOCATE( zcfield3d , zcmask )
 

   END SUBROUTINE crsfun_TW

SUBROUTINE crs_e3_max( p_e3, cd_type, p_mask, p_e3_crs)
      !!----------------------------------------------------------------
      !!               *** SUBROUTINE crsfun_TW ***
      !! ** Purpose :  Five applications.
      !!               1) Maximum surface quantity 
      !!                  - Vertical scale factors (fse3t or fse3w) 
      !!                    max thickness of the parent grid for coarse grid scale factors.
      !!                  - or diffusion test
      !!               2) Area-weighted mean quantity: w, or other 3D or 2D quantity
      !!               3) Volume-weighted mean quantity: tracer
      !!               4) Minimum surface quantity (diffusion test)
      !!               5) Area- or Volume- weighted sum.
      !! ** Method  :  1) - cd_op = 'MAX'. Determines the max vertical thickness of grid boxes
      !!                    including partial steps for at the bottom
      !!                    for the coarsened grid, where within the subset of 
      !!                    the parent grid cells the maximum thickness is taken.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!                    Where, normally p_pfield3d_1 is e3t.
      !!                  - cd_op = 'MAX'. May also be used for say, determining the maximum value of Kz, 
      !!                    thus p_pfield3d_1 is set to the 3D field, Kz.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!               2) - cd_op = 'ARE'. Calculate the area-weighted average (surface e1t*e2t)  
      !!                    of vertical velocity, w.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!                  - cd_op = 'ARE'. Calculate area-weighted average of a 2D quantity (e.g. emp)
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield2d
      !!               3) - cd_op = 'VOL'. Calculate the ocean volume (e1e2t*[fse3t|fse3w]) 
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!                  - cd_op = 'VOL'. Calculate volume-weighted average (volume e1t*e2t*fse3t) of a quantity.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1, p_pfield3d_2
      !!               4) - cd_op = 'MIN'. Calculate the minimum value on surface e1t*e2t for 3D variables
      !!                  Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!               5) - cd_op = 'SUM'. Calculate a dimesionally-weighted sum.  This could be area-weighted
      !!                  or volume-weighted sum. 
      !! ** Inputs  : p_e1e2t      = parent grid top face surface area, e1t*e2t 
      !!              cd_type      = grid type T, W (U, V, F) 
      !!              cd_op        = MAX, ARE, VOL, MIN, SUM
      !!              p_cmask      =  coarse grid mask
      !!              p_ptmask     =  parent grid tmask     
      !!              psgn         = (Optional) sign for lbc_lnk  
      !!              p_pfield2d   = (Optional) 2D field on parent grid
      !!              p_pfield3d_1 = (Optional) parent grid fse3t or fse3w
      !!              p_pfield3d_2 = (Optional) 3D field on parent grid
      !! ** Outputs : p_cfield2d   = (Optional) 2D field on coarse grid
      !!              p_cfield3d   = (Optional) 3D field on coarse grid
      !!
      !! 
      !! History.  30 May.  Editing.  To decide later: Keep all functionality or separate out the mean function.
      !!            7 Jun   TODO. Need to fix up the parent grid mask to optional like crsfun_UV...
      !!----------------------------------------------------------------
      !! 
      !!  Arguments
      CHARACTER(len=1),                           INTENT(in) :: cd_type      ! grid type T, W ( U, V, F)
      REAL(wp), DIMENSION(jpi,jpj,jpk),           INTENT(in) :: p_mask       ! Parent grid T mask
      REAL(wp), DIMENSION(jpi,jpj,jpk), OPTIONAL, INTENT(in) :: p_e3         ! 3D tracer T or W on parent grid
      REAL(wp), DIMENSION(jpi_crs,jpj_crs,jpk), OPTIONAL, INTENT(out):: p_e3_crs ! Coarse grid box east or north face quantity 

      !! Local variables
      INTEGER ::  ji, jj, jk                   ! dummy loop indices
      INTEGER :: ijie,ijis,ijje,ijjs,ijpk,jii,jjj
      REAL(wp), ALLOCATABLE, DIMENSION(:,:,:) :: ze3, ze3_crs, zpmask  
      !!----------------------------------------------------------------  
      ! Initialize

      ALLOCATE( ze3(jpi,jpj,jpk), zpmask(jpi,jpj,jpk) )
      ALLOCATE( ze3_crs(jpi_crs,jpj_crs,jpk) )

      ! Arrays, scalars initialization 
      ze3(:,:,:)       = p_e3(:,:,:)
      ze3_crs(:,:,:)   = 0.0
      zpmask(:,:,:)    = p_mask(:,:,:)
      ijpk             = jpk

      SELECT CASE ( cd_type )
      
         CASE ('T')
         
            DO jk = 1 , ijpk
            
               DO ji = 1, jpi_crs                       ! ji = 1 et jpi_crs definit par cyclique est-ouest et pivot T
                   ijie = mie_crs(ji)
                   ijis = mis_crs(ji)

                   DO jj = 1, jpj_crs                   ! jj = jpj_crs definit par pivot T 
                      ijje = mje_crs(jj) 
                      ijjs = mjs_crs(jj)  
                   
                      DO jii = ijis, ijie
                         DO jjj = ijjs, ijje
                            ze3_crs(ji,jj,jk) = max( ze3_crs(ji,jj,jk), ze3(jii,jjj,jk) * zpmask(jii,jjj,jk)  )
                         ENDDO
                      ENDDO
                   ENDDO
               ENDDO
            ENDDO
            
         CASE ('W')
         
            DO jk = 2 , ijpk
            
               DO ji = 1, jpi_crs                       ! ji = 1 et jpi_crs definit par cyclique est-ouest et pivot T
                   ijie = mie_crs(ji)
                   ijis = mis_crs(ji)

                   DO jj = 1, jpj_crs                   ! jj = jpj_crs definit par pivot T 
                      ijje = mje_crs(jj) 
                      ijjs = mjs_crs(jj)  
                   
                      DO jii = ijis, ijie
                         DO jjj = ijjs, ijje
                            ze3_crs(ji,jj,jk) = max( ze3_crs(ji,jj,jk), ze3(jii,jjj,jk) * zpmask(jii,jjj,jk-1)  )
                         ENDDO
                      ENDDO
                   ENDDO
               ENDDO
            ENDDO
             
            jk = 1                                           ! cas particulier car zpmask(jii,jjj,0) n'existe pas      
          
            DO ji = 1, jpi_crs                       
               ijie = mie_crs(ji)
               ijis = mis_crs(ji)

               DO jj = 1, jpj_crs                 
                  ijje = mje_crs(jj) 
                  ijjs = mjs_crs(jj)  
                   
                  DO jii = ijis, ijie
                     DO jjj = ijjs, ijje
                        ze3_crs(ji,jj,jk) = max( ze3_crs(ji,jj,jk), ze3(jii,jjj,jk) * zpmask(jii,jjj,jk)  )
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO  
            
         END SELECT 
         
         p_e3_crs(:,:,:) = ze3_crs(:,:,:)
         
         CALL crs_lbc_lnk( cd_type, 1.0, pt3d1=p_e3_crs )

         ! lbcnlk met la ligne jpj = 1 a 0 donc il faut la remettre en ne pas oubliant le cyclique est-ouest
                   
         p_e3_crs(   :   ,1,:) = ze3_crs(    :    ,1,:)   
         p_e3_crs(   1   ,1,:) = ze3_crs(jpi_crsm1,1,:)
         p_e3_crs(jpi_crs,1,:) = ze3_crs(    2    ,1,:)       
 

      DEALLOCATE( ze3 , zpmask )
      DEALLOCATE( ze3_crs )
 

   END SUBROUTINE crs_e3_max


SUBROUTINE crs_surf(p_e1, p_e2 ,p_e3, cd_type, p_mask, surf_crs, surf_msk_crs)
      !!----------------------------------------------------------------
      !!               *** SUBROUTINE crsfun_TW ***
      !! ** Purpose :  Five applications.
      !!               1) Maximum surface quantity 
      !!                  - Vertical scale factors (fse3t or fse3w) 
      !!                    max thickness of the parent grid for coarse grid scale factors.
      !!                  - or diffusion test
      !!               2) Area-weighted mean quantity: w, or other 3D or 2D quantity
      !!               3) Volume-weighted mean quantity: tracer
      !!               4) Minimum surface quantity (diffusion test)
      !!               5) Area- or Volume- weighted sum.
      !! ** Method  :  1) - cd_op = 'MAX'. Determines the max vertical thickness of grid boxes
      !!                    including partial steps for at the bottom
      !!                    for the coarsened grid, where within the subset of 
      !!                    the parent grid cells the maximum thickness is taken.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!                    Where, normally p_pfield3d_1 is e3t.
      !!                  - cd_op = 'MAX'. May also be used for say, determining the maximum value of Kz, 
      !!                    thus p_pfield3d_1 is set to the 3D field, Kz.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!               2) - cd_op = 'ARE'. Calculate the area-weighted average (surface e1t*e2t)  
      !!                    of vertical velocity, w.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!                  - cd_op = 'ARE'. Calculate area-weighted average of a 2D quantity (e.g. emp)
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield2d
      !!               3) - cd_op = 'VOL'. Calculate the ocean volume (e1e2t*[fse3t|fse3w]) 
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!                  - cd_op = 'VOL'. Calculate volume-weighted average (volume e1t*e2t*fse3t) of a quantity.
      !!                    Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1, p_pfield3d_2
      !!               4) - cd_op = 'MIN'. Calculate the minimum value on surface e1t*e2t for 3D variables
      !!                  Valid arguments: p_e1e2t, cd_type, cd_op, p_cmask, p_pfield3d_1
      !!               5) - cd_op = 'SUM'. Calculate a dimesionally-weighted sum.  This could be area-weighted
      !!                  or volume-weighted sum. 
      !! ** Inputs  : p_e1e2t      = parent grid top face surface area, e1t*e2t 
      !!              cd_type      = grid type T, W (U, V, F) 
      !!              cd_op        = MAX, ARE, VOL, MIN, SUM
      !!              p_cmask      =  coarse grid mask
      !!              p_ptmask     =  parent grid tmask     
      !!              psgn         = (Optional) sign for lbc_lnk  
      !!              p_pfield2d   = (Optional) 2D field on parent grid
      !!              p_pfield3d_1 = (Optional) parent grid fse3t or fse3w
      !!              p_pfield3d_2 = (Optional) 3D field on parent grid
      !! ** Outputs : p_cfield2d   = (Optional) 2D field on coarse grid
      !!              p_cfield3d   = (Optional) 3D field on coarse grid
      !!
      !! 
      !! History.  30 May.  Editing.  To decide later: Keep all functionality or separate out the mean function.
      !!            7 Jun   TODO. Need to fix up the parent grid mask to optional like crsfun_UV...
      !!----------------------------------------------------------------
      !! 
      !!  Arguments
      CHARACTER(len=1),                           INTENT(in) :: cd_type      ! grid type T, W ( U, V, F)
      REAL(wp), DIMENSION(jpi,jpj,jpk),           INTENT(in) :: p_mask       ! Parent grid T mask
      REAL(wp), DIMENSION(jpi,jpj,jpk), OPTIONAL, INTENT(in) :: p_e1, p_e2, p_e3         ! 3D tracer T or W on parent grid
      REAL(wp), DIMENSION(jpi_crs,jpj_crs,jpk), OPTIONAL, INTENT(out):: surf_crs, surf_msk_crs ! Coarse grid box east or north face quantity 

      !! Local variables
      INTEGER ::  ji, jj, jk                   ! dummy loop indices
      INTEGER :: ijie,ijis,ijje,ijjs,ijpk,jii,jjj
      REAL(wp), ALLOCATABLE, DIMENSION(:,:,:) :: ze1, ze2, ze3, zsurf_crs, zsurf_msk_crs, zpmask  
      !!----------------------------------------------------------------  
      ! Initialize

      ALLOCATE( ze1(jpi,jpj,jpk), ze2(jpi,jpj,jpk), ze3(jpi,jpj,jpk), zpmask(jpi,jpj,jpk) )
      ALLOCATE( zsurf_crs(jpi_crs,jpj_crs,jpk), zsurf_msk_crs(jpi_crs,jpj_crs,jpk) )

      ! Arrays, scalars initialization 
      ze1(:,:,:)           = p_e1(:,:,:)
      ze2(:,:,:)           = p_e2(:,:,:)
      ze3(:,:,:)           = p_e3(:,:,:)
      zsurf_crs(:,:,:)     = 0.0
      zsurf_msk_crs(:,:,:) = 0.0
      zpmask(:,:,:)        = p_mask(:,:,:)
      ijpk                 = jpk

      SELECT CASE ( cd_type )
      
         CASE ('W')

            DO jk = 2 , ijpk
            
               DO ji = 1, jpi_crs                       ! ji = 1 et jpi_crs definit par cyclique est-ouest et pivot T
                   ijie = mie_crs(ji)
                   ijis = mis_crs(ji)
                   jj   = 1
                   ijje = mje_crs(jj) 
                   ijjs = mjs_crs(jj)
                   
                   DO jii = ijis, ijie
                      DO jjj = ijjs, ijje
                         zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) + ze1(ji,jj,jk) * ze2(jii,jjj,jk)
                         zsurf_msk_crs(ji,jj,jk) = zsurf_msk_crs(ji,jj,jk) + ( ze1(ji,jj,jk) * ze2(jii,jjj,jk) ) * zpmask(jii,jjj,jk-1)               
                      ENDDO
                   ENDDO
                   
                   zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) * 3
                   
                   DO jj = 2, jpj_crs                   ! jj = jpj_crs definit par pivot T 
                      ijje = mje_crs(jj) 
                      ijjs = mjs_crs(jj)  
                   
                      DO jii = ijis, ijie
                         DO jjj = ijjs, ijje
                            zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) + ze1(ji,jj,jk) * ze2(jii,jjj,jk)
                            zsurf_msk_crs(ji,jj,jk) = zsurf_msk_crs(ji,jj,jk) + ( ze1(ji,jj,jk) * ze2(jii,jjj,jk) ) * zpmask(jii,jjj,jk-1)   
                         ENDDO
                      ENDDO
                   ENDDO
               ENDDO
            ENDDO
            
            jk = 1                                      !cas particulier ou on est en surface
            
            DO ji = 1, jpi_crs                       ! ji = 1 et jpi_crs definit par cyclique est-ouest et pivot T
               ijie = mie_crs(ji)
               ijis = mis_crs(ji)
               jj   = 1                           
               ijje = mje_crs(jj) 
               ijjs = mjs_crs(jj)
                   
               DO jii = ijis, ijie 
                  DO jjj = ijjs, ijje
                     zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) + ze1(ji,jj,jk) * ze2(jii,jjj,jk)
                     zsurf_msk_crs(ji,jj,jk) = zsurf_msk_crs(ji,jj,jk) + ( ze1(ji,jj,jk) * ze2(jii,jjj,jk) ) * zpmask(jii,jjj,jk)               
                  ENDDO  
               ENDDO
               zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) * 3                
               DO jj = 2, jpj_crs                   ! jj = jpj_crs definit par pivot T 
                  ijje = mje_crs(jj) 
                  ijjs = mjs_crs(jj)              
                  DO jii = ijis, ijie
                     DO jjj = ijjs, ijje   
                        zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) + ze1(ji,jj,jk) * ze2(jii,jjj,jk)
                        zsurf_msk_crs(ji,jj,jk) = zsurf_msk_crs(ji,jj,jk) + ( ze1(ji,jj,jk) * ze2(jii,jjj,jk) ) * zpmask(jii,jjj,jk)   
                     ENDDO   
                  ENDDO
               ENDDO
            ENDDO
            
       CASE ('U')
       
          DO jk = 1 , ijpk
            
             DO ji = 1, jpi_crs                       ! ji = 1 et jpi_crs definit par cyclique est-ouest et pivot T
                ijie = mie_crs(ji)
                ijis = mis_crs(ji)
                jj   = 1
                ijje = mje_crs(jj)  
                ijjs = mjs_crs(jj)
                   
                DO jii = ijis, ijie
                   DO jjj = ijjs, ijje      
                      zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) + ze3(ji,jj,jk) * ze2(jii,jjj,jk)
                      zsurf_msk_crs(ji,jj,jk) = zsurf_msk_crs(ji,jj,jk) + ( ze3(ji,jj,jk) * ze2(jii,jjj,jk) ) * zpmask(jii,jjj,jk)               
                   ENDDO
                ENDDO
                   
                zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) * 3
                   
                DO jj = 2, jpj_crs                   ! jj = jpj_crs definit par pivot T 
                   ijje = mje_crs(jj) 
                   ijjs = mjs_crs(jj)  
               
                   DO jii = ijis, ijie
                      DO jjj = ijjs, ijje
                         zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) + ze3(ji,jj,jk) * ze2(jii,jjj,jk)
                         zsurf_msk_crs(ji,jj,jk) = zsurf_msk_crs(ji,jj,jk) + ( ze3(ji,jj,jk) * ze2(jii,jjj,jk) ) * zpmask(jii,jjj,jk)
                      ENDDO         
                   ENDDO
                ENDDO
            ENDDO
         ENDDO
       
      CASE ('V')
          
         DO jk = 1 , ijpk
            
            DO ji = 1, jpi_crs                       ! ji = 1 et jpi_crs definit par cyclique est-ouest et pivot T
               ijie = mie_crs(ji)
               ijis = mis_crs(ji)
                   
               DO jj = 1, jpj_crs                   ! jj = jpj_crs definit par pivot T 
                  ijje = mje_crs(jj)   
                  ijjs = mjs_crs(jj)  
                   
                  DO jii = ijis, ijie
                     DO jjj = ijjs, ijje
                        zsurf_crs(ji,jj,jk) = zsurf_crs(ji,jj,jk) + ze3(ji,jj,jk) * ze1(jii,jjj,jk) 
                        zsurf_msk_crs(ji,jj,jk) = zsurf_msk_crs(ji,jj,jk) + ( ze3(ji,jj,jk) * ze1(jii,jjj,jk) ) * zpmask(jii,jjj,jk)   
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      END SELECT
 
      surf_crs(:,:,:) = zsurf_crs(:,:,:)
      CALL crs_lbc_lnk( cd_type, 1.0, pt3d1=surf_crs )
         ! lbcnlk met la ligne jpj = 1 a 0 donc il faut la remettre en ne pas oubliant le cyclique est-ouest 
         ! a faire un case pour cyclique est-ouest ou non.      
         surf_crs(   :   ,1,:) = zsurf_crs(    :    ,1,:)   
         surf_crs(   1   ,1,:) = zsurf_crs(jpi_crsm1,1,:)
         surf_crs(jpi_crs,1,:) = zsurf_crs(    2    ,1,:)   
         
         surf_msk_crs(:,:,:) = zsurf_msk_crs(:,:,:)
      CALL crs_lbc_lnk( cd_type, 1.0, pt3d1=surf_msk_crs )
         ! lbcnlk met la ligne jpj = 1 a 0 donc il faut la remettre en ne pas oubliant le cyclique est-ouest        
         surf_msk_crs(   :   ,1,:) = zsurf_msk_crs(    :    ,1,:)   
         surf_msk_crs(   1   ,1,:) = zsurf_msk_crs(jpi_crsm1,1,:)
         surf_msk_crs(jpi_crs,1,:) = zsurf_msk_crs(    2    ,1,:) 

      DEALLOCATE( ze3 , ze2, ze1, zpmask )
      DEALLOCATE( zsurf_msk_crs, zsurf_crs )
 

   END SUBROUTINE crs_surf


END MODULE crs
